@@ -76,6 +76,48 @@ struct TaggedStatement {
     const char* text;
 };
 
+/// @brief Retry on InnoDB deadlock.
+///
+/// When f(args) returns ER_LOCK_DEADLOCK f(args) is called again up to 5 times.
+///
+/// @tparam Fun Type of the function which must return an int.
+/// @tparam Args Types of arguments.
+/// @param fun The function to call.
+/// @param args Arguments.
+/// @return status (can be ER_LOCK_DEADLOCK after 5 retries).
+template <typename Fun, typename... Args>
+int retryOnDeadlock(Fun& fun, Args... args) {
+    int status;
+    for (unsigned count = 0; count < 5; ++count) {
+        status = fun(args...);
+        if (status != ER_LOCK_DEADLOCK) {
+            break;
+        }
+    }
+    return (status);
+}
+
+/// @brief Execute a prepared statement.
+///
+/// Call mysql_stmt_execute and retry on ER_LOCK_DEADLOCK.
+///
+/// @param stmt Statement to execute.
+/// @return status (can be ER_LOCK_DEADLOCK after 5 retries).
+inline int MysqlExecuteStatement(MYSQL_STMT* stmt) {
+    return (retryOnDeadlock(mysql_stmt_execute, stmt));
+}
+
+/// @brief Execute a literal statement.
+///
+/// Call mysql_query and retry on ER_LOCK_DEADLOCK.
+///
+/// @param mysql MySQL context.
+/// @param stmt Statement to execute.
+/// @return status (can be ER_LOCK_DEADLOCK after 5 retries).
+inline int MysqlQuery(MYSQL* mysql, const char* stmt) {
+    return (retryOnDeadlock(mysql_query, mysql, stmt));
+}
+
 /// @brief MySQL Handle Holder
 ///
 /// Small RAII object for safer initialization, will close the database
@@ -357,6 +399,7 @@ public:
                      const MySqlBindingCollection& in_bindings,
                      MySqlBindingCollection& out_bindings,
                      ConsumeResultFun process_result) {
+        checkUnusable();
         // Extract native input bindings.
         std::vector<MYSQL_BIND> in_bind_vec;
         for (MySqlBindingPtr in_binding : in_bindings) {
@@ -382,7 +425,7 @@ public:
         }
 
         // Execute query.
-        status = mysql_stmt_execute(statements_[index]);
+        status = MysqlExecuteStatement(statements_[index]);
         checkError(status, index, "unable to execute");
 
         status = mysql_stmt_store_result(statements_[index]);
@@ -434,6 +477,7 @@ public:
     template<typename StatementIndex>
     void insertQuery(const StatementIndex& index,
                      const MySqlBindingCollection& in_bindings) {
+        checkUnusable();
         std::vector<MYSQL_BIND> in_bind_vec;
         for (MySqlBindingPtr in_binding : in_bindings) {
             in_bind_vec.push_back(in_binding->getMySqlBinding());
@@ -445,7 +489,7 @@ public:
         checkError(status, index, "unable to bind parameters");
 
         // Execute the statement
-        status = mysql_stmt_execute(statements_[index]);
+        status = MysqlExecuteStatement(statements_[index]);
 
         if (status != 0) {
             // Failure: check for the special case of duplicate entry.
@@ -477,6 +521,7 @@ public:
     template<typename StatementIndex>
     uint64_t updateDeleteQuery(const StatementIndex& index,
                                const MySqlBindingCollection& in_bindings) {
+        checkUnusable();
         std::vector<MYSQL_BIND> in_bind_vec;
         for (MySqlBindingPtr in_binding : in_bindings) {
             in_bind_vec.push_back(in_binding->getMySqlBinding());
@@ -488,7 +533,7 @@ public:
         checkError(status, index, "unable to bind parameters");
 
         // Execute the statement
-        status = mysql_stmt_execute(statements_[index]);
+        status = MysqlExecuteStatement(statements_[index]);
 
         if (status != 0) {
             // Failure: check for the special case of duplicate entry.
@@ -561,7 +606,7 @@ public:
     ///        failed.
     template<typename StatementIndex>
     void checkError(const int status, const StatementIndex& index,
-                    const char* what) const {
+                    const char* what) {
         if (status != 0) {
             switch(mysql_errno(mysql_)) {
                 // These are the ones we consider fatal. Remember this method is
@@ -580,6 +625,9 @@ public:
                     .arg(mysql_error(mysql_))
                     .arg(mysql_errno(mysql_));
 
+                // Mark the connection as no longer usuable.
+                markUnusable();
+
                 // If there's no lost db callback or it returns false,
                 // then we're not attempting to recover so we're done.
                 if (!invokeDbLostCallback()) {
@@ -589,7 +637,7 @@ public:
 
                 // We still need to throw so caller can error out of the current
                 // processing.
-                isc_throw(db::DbOperationError,
+                isc_throw(db::DbConnectionUnusable,
                           "fatal database errror or connectivity lost");
             default:
                 // Connection is ok, so it must be an SQL error
